from dune.perftool.generation import (backend,
                                      kernel_cached,
                                      valuearg, temporary_variable, instruction)
from dune.perftool.options import get_option
from dune.perftool.pdelab.argument import CoefficientAccess
from dune.perftool.blockstructured.tools import micro_index_to_macro_index, sub_element_inames
from loopy.types import NumpyType
import pymbolic.primitives as prim
import loopy as lp


def define_alias(container, lfs, element):
    k = get_option("number_of_blocks")
    p = element.degree()
    temporary_variable(container+"_alias", shape=(k + 1, k + 1, p + 1, p + 1), strides=(p, p * k + 1, 1, k + 1),
                       base_storage="dummy_"+container, managed=True, _base_storage_access_may_be_aliasing=True)
    code = "{} = &{}({},0);".format(container+"_alias",container, lfs.name)
    instruction(within_inames=frozenset(),
                code=code,
                read_variables=frozenset({container}),
                assignees=frozenset({container+"_alias"}))


def name_alias(container, lfs, element):
    define_alias(container, lfs, element)
    name = container+"_alias"
    return name


# TODO remove the need for element
@kernel_cached
@backend(interface="pymbolic_coefficient", name="blockstructured")
def pymbolic_coefficient(container, lfs, element, index):
    # TODO introduce a proper type for local function spaces!
    if isinstance(lfs, str):
        valuearg(lfs, dtype=NumpyType("str"))

    # If the LFS is not yet a pymbolic expression, make it one
    if not isinstance(lfs, prim.Expression):
        lfs = prim.Variable(lfs)

    # use higher order FEM index instead of Q1 index
    if get_option("vectorization_blockstructured"):
        subelem_inames = sub_element_inames()
        coeff_alias = name_alias(container, lfs, element)
        return prim.Subscript(lp.TaggedVariable(coeff_alias, 'coeff_alias'), tuple(prim.Variable(i) for i in subelem_inames+index))
    else:
        return prim.Call(CoefficientAccess(container), (lfs, micro_index_to_macro_index(element, index),))
